#lang racket

; interpreter   -------------------------------------------------------

(define run
  (lambda (str)
    (let ((program (parse str))
          (stack  '()))
      (letrec ((eval (lambda (pgm stk)
                     (if (null? pgm)
                         stk
                         (eval (rest pgm)
                               (eval-exp (first pgm) stk))))))
        (eval program stack)))))

;; a simple demo program

(define p "4 5 + dup *")  ; (run p)

; parser   ------------------------------------------------------------

(define parse
  (lambda (str)
    (letrec ((parse-list
                (lambda (lst-of-str)
                  (if (null? lst-of-str)
                      '()
                      (let ((first (car lst-of-str))
                            (rest  (cdr lst-of-str)))
                        (let ((entry (assoc first symbol-table)))
                          (cons (if entry
                                    (cdr entry)
                                    (string->number first))
                                (parse-list rest))))))))
      (parse-list (string-split str)))))

(define symbol-table
  '( ("+"      . add)
     ("-"      . sub)
     ("*"      . mul)
     ("/"      . div)
     ("inc"    . inc)
     ("dec"    . dec)
     ("sqrt"   . sqrt)
     ("abs"    . abs)
     ("drop"   . drop)
     ("dup"    . dup)
     ("swap"   . swap)
     ("rollup" . rollup) ))

; evaluator   ---------------------------------------------------------

(define eval-exp
  (lambda (val stk)
    (cond ((number?    val) (cons val stk))
          ((unary-op?  val) (eval-unary val stk))
          ((binary-op? val) (eval-binary val stk))
          ((stack-op?  val) (eval-stackop val stk))
          (else (error 'eval "illegal value ~a" val)))))

(define eval-unary
  (lambda (val stk)
    (cond ((null? stk)     (error 'eval-unary "stack underflow"))
          ((eq? 'inc  val) (cons (add1 (car stk)) (cdr stk)))
          ((eq? 'dec  val) (cons (sub1 (car stk)) (cdr stk)))
          ((eq? 'sqrt val) (cons (sqrt (car stk)) (cdr stk)))
          ((eq? 'abs  val) (cons (abs  (car stk)) (cdr stk)))
          (else (error 'eval-unary "illegal operator ~a" val)))))

(define eval-binary
  (lambda (val stk)
    (if (or (null? stk) (null? (cdr stk)))
        (error 'eval-binary "stack underflow")
        (let ((arg2 (car  stk))
              (arg1 (cadr stk))
              (rest (cddr stk)))
          (cond ((eq? 'add  val) (cons (+ arg1 arg2) rest))
                ((eq? 'sub  val) (cons (- arg1 arg2) rest))
                ((eq? 'mul  val) (cons (* arg1 arg2) rest))
                ((eq? 'div  val) (cons (/ arg1 arg2) rest))
                (else (error 'eval-binary "illegal operator ~a" val)))))))

(define eval-stackop
  (lambda (val stk)
    (cond ((eq? 'drop val) (cdr stk))
          ((eq? 'dup  val) (cons (car stk) stk))
          ((eq? 'swap val) (list* (cadr stk) (car stk) (cddr stk)))
          ((eq? 'rollup val)
              (let ((1st (car stk))
                    (2nd (cadr stk))
                    (3rd (caddr stk))
                    (rst (cdddr stk)))
                (list* 2nd 3rd 1st rst)))
          (else (error 'eval-stackop "illegal operator ~a" val)))))

; syntax procs   ------------------------------------------------------
;                There isn't much syntax, so we don't need much in the
;                the way of syntax procedures.  These type predicates
;                help the evaluator choose its cases.

(define binary-op?
  (lambda (x)
    (member x '(add sub mul div))))

(define unary-op?
  (lambda (x)
    (member x '(inc dec sqrt abs))))

(define stack-op?
  (lambda (x)
    (member x '(drop dup swap rollup))))

(define exp?
  (lambda (x)
    (or (number? x)
        (unary-op? x)
        (binary-op? x)
        (stack-op? x))))

; end   ---------------------------------------------------------------

